<?php
/**
 * Class for managing ILS-specific authentication.
 *
 * PHP version 7
 *
 * Copyright (C) Villanova University 2007.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  Authentication
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Page
 */
namespace finc\Auth;

use VuFind\Auth\EmailAuthenticator as EmailAuthenticator;
use VuFind\Auth\Manager as Manager;
use VuFind\ILS\Connection as ILSConnection;
use Laminas\Session\SessionManager as SessionManager;

/**
 * Class for managing ILS-specific authentication.
 *
 * @category VuFind
 * @package  Authentication
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Page
 */
class ILSAuthenticator extends \VuFind\Auth\ILSAuthenticator
{
    /**
     * Session containing PAIA login information
     *
     * @var Laminas\Session\Container
     */
    protected $session;

    /**
     * SessionManager
     *
     * @var \VuFind\SessionManager
     */
    protected $sessionManager;

    /**
     * Constructor
     *
     * @param Manager            $auth           Auth manager
     * @param ILSConnection      $catalog        ILS connection
     * @param SessionManager     $sessionManager Session manager
     * @param EmailAuthenticator $emailAuth      Email authenticator
     */
    public function __construct(
        Manager $auth,
        ILSConnection $catalog,
        SessionManager $sessionManager,
        EmailAuthenticator $emailAuth = null
    ) {
        $this->auth = $auth;
        $this->catalog = $catalog;
        $this->sessionManager = $sessionManager;
        $this->emailAuthenticator = $emailAuth;
    }

    /**
     * Get the session container (constructing it on demand if not already present)
     *
     * @return SessionContainer
     */
    protected function getSession()
    {
        // SessionContainer not defined yet? Build it now:
        if (null === $this->session) {
            $this->session = new \Laminas\Session\Container(
                'PAIA',
                $this->sessionManager
            );
        }
        return $this->session;
    }

    /**
     * Log the current user into the catalog using stored credentials; if this
     * fails, clear the user's stored credentials so they can enter new, corrected
     * ones.
     *
     * Returns associative array of patron data on success, false on failure.
     *
     * @return array|bool
     */
    public function storedCatalogLogin()
    {
        // Fail if no username is found, but allow a missing password (not every ILS
        // requires a password to connect).
        if (($user = $this->auth->isLoggedIn()) && !empty($user->cat_username)) {
            // Do we have a previously cached ILS account?
            if (isset($this->ilsAccount[$user->cat_username])) {
                return $this->ilsAccount[$user->cat_username];
            }

            $session = $this->getSession();
            if (isset($session->cat_password)) {
                $patron = $this->catalog->patronLogin(
                    $user->cat_username,
                    $session->cat_password
                );
            } else {
                $patron = $this->catalog->patronLogin(
                    $user->cat_username,
                    $user->getCatPassword()
                );
            }

            if (empty($patron)) {
                // Problem logging in -- clear user credentials so they can be
                // prompted again; perhaps their password has changed in the
                // system!
                $user->clearCredentials();
            } else {
                // cache for future use
                $this->ilsAccount[$user->cat_username] = $patron;
                return $patron;
            }
        }

        return false;
    }

    /**
     * Update current user account with the patron information
     *
     * @param string $catUsername Catalog username
     * @param string $catPassword Catalog password
     * @param array  $patron      Patron
     *
     * @return void
     */
    protected function updateUser($catUsername, $catPassword, $patron)
    {
        $user = $this->auth->isLoggedIn();
        if ($user) {
            $user->saveCredentials($catUsername, null);
            $this->auth->updateSession($user);
            // cache for future use
            $this->ilsAccount[$catUsername] = $patron;
        }
    }
}
