<?php
/**
 * Copyright (C) Leipzig University Library 2020.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @author   Sebastian Kehr <kehr@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development Wiki
 */
namespace finc\Controller\Admin;

use Symfony\Component\Filesystem\Filesystem as Fs;
use VuFind\Auth\InvalidArgumentException;
use VuFind\Config\Writer as ConfigWriter;
use VuFind\I18n\Translator\Loader\ExtendedIni;
use VuFindAdmin\Controller\AbstractAdmin;
use Laminas\I18n\Translator\Translator;
use Laminas\ServiceManager\ServiceLocatorInterface;

class I18nController extends AbstractAdmin
{
    /**
     * @var string[]
     */
    protected $dirs;

    /**
     * @var Translator
     */
    protected $translator;

    /**
     * @var \ReflectionProperty[]
     */
    protected $properties;

    /**
     * @var string[]
     */
    protected $languages;

    /**
     * @var array
     */
    protected $defaultTranslations;

    /**
     * @var array
     */
    protected $translations;

    /**
     * @var string
     */
    private $locale;

    /**
     * I18nController constructor.
     *
     * @param ServiceLocatorInterface $sl
     * @param Translator              $translator
     * @param array                   $languages
     */
    public function __construct(
        ServiceLocatorInterface $sl,
        Translator $translator,
        array $languages
    ) {
        parent::__construct($sl);
        $this->translator = $translator;
        $this->dirs = $this->getDirs();
        $languages = array_map(
            function ($name, $locale) {
                $name = $this->translator->translate($name);
                return compact('name', 'locale');
            },
            $languages,
            $locales = array_keys($languages)
        );
        $this->languages = array_combine($locales, $languages);
    }

    public function homeAction()
    {
        $params = $this->params();
        $defaultLocale = array_keys($this->languages)[0];
        $this->locale = $locale = $params->fromQuery('locale', $defaultLocale);
        $domain = $params->fromQuery('domain', 'default');

        $translations = $this->getTranslations();
        $defaultTranslations = $this->getDefaultTranslations();

        foreach (array_keys($translations[$this->locale]) as $name) {
            $selected = $name === $domain;
            $domains[$name] = compact('name', 'selected');
        }

        $languages = array_map(
            function ($lang) {
                $langLocale = $lang['locale'];
                $selected = $langLocale === $this->locale;
                return $lang + compact('selected');
            },
            $this->languages
        );

        $view = $this->createViewModel();
        $view->setTemplate('admin/i18n/home');
        $view->setVariables(
            compact(
                'domains',
                'languages',
                'translations',
                'defaultTranslations',
                'domain',
                'locale'
            )
        );

        return $view;
    }

    public function editAction()
    {
        $view = $this->createViewModel();
        $view->setTemplate('admin/i18n/edit');
        $params = $this->params();
        $view->setVariables(
            [
                'locale' => $locale = $params->fromPost('locale'),
                'language' => $this->languages[$locale]['name'],
                'domain' => $params->fromPost('domain'),
                'token' => $params->fromPost('token'),
                'value' => $params->fromPost('value'),
                'defaultValue' => $params->fromPost('default_value')
            ]
        );

        return $view;
    }

    public function saveAction()
    {
        $params = $this->params();
        $domain = $params->fromPost('domain');
        $value = $params->fromPost('value');
        $defaultValue = $params->fromPost('default_value');
        $token = $params->fromPost('token');
        $locale = $params->fromPost('locale');
        $reset = !!$params->fromPost('reset') || $value === $defaultValue;

        return $this->saveTranslation(
            $domain,
            $token,
            $locale,
            $reset ? null : $value
        );
    }

    protected function getTranslations(): array
    {
        if (!$this->translations) {
            $this->translations = $this->loadTranslations($this->dirs);
        }

        return $this->translations;
    }

    protected function getDefaultTranslations(): array
    {
        if (!$this->defaultTranslations) {
            $dirs = array_slice($this->dirs, 0, -1);
            $this->defaultTranslations = $this->loadTranslations($dirs);
        }

        return $this->defaultTranslations;
    }

    protected function loadTranslations(array $dirs)
    {
        foreach ($this->getDomains() as $domain) {
            $translations[$this->locale][$domain] = $this->loadMessages(
                $dirs,
                $domain,
                $this->locale
            );
        }

        return $translations ?? [];
    }

    protected function loadMessages(
        array $dirs,
        string $domain,
        string $locale
    ) {
        $translator = new Translator();
        $plugin = $this->translator->getPluginManager()->get('ExtendedIni');
        $locales = $this->getPropertyValue($plugin, 'fallbackLocales');
        $reader = $this->getPropertyValue($plugin, 'reader');
        $service = new ExtendedIni($dirs, $locales, $reader);
        $translator->getPluginManager()->setService('ExtendedIni', $service);
        $translator->addTranslationFile(
            'ExtendedIni',
            $domain === 'default' ? null : $domain,
            $domain,
            $locale
        );
        $messages = $translator->getAllMessages($domain, $locale);
        foreach ($messages->getArrayCopy() as $key => $message) {
            if ($key[0] === '#' || $key === '@parent_ini') {
                unset($messages[$key]);
            }
        }

        return $messages;
    }

    protected function getDomains()
    {
        return array_keys($this->getPropertyValue($this->translator, 'files'));
    }

    protected function getDirs(): array
    {
        $object = $this->translator->getPluginManager()->get('ExtendedIni');
        return $this->getPropertyValue($object, 'pathStack');
    }

    protected function saveTranslation($domain, $token, $locale, $value)
    {
        $fs = new FS();
        list($filename) = array_slice($this->dirs, -1);
        if ($domain !== 'default') {
            $filename .= "/$domain";
        }
        $filename .= "/$locale.ini";

        try {
            if (empty($token)) {
                throw new InvalidArgumentException(
                    $this->translate('This field is required') . ": Token"
                );
            }

            if (!$this->isValidLocal($locale)) {
                throw new InvalidArgumentException(
                    $this->translate('permission_denied_title')
                );
            }

            $fs->mkdir(dirname($filename));
            $fs->touch($filename);

            $writer = new ConfigWriter($filename);
            $writer->set(null, $token, $value);
            file_put_contents($filename, trim($writer->getContent()) . "\n");
            $cacheId = $this->translator->getCacheId($domain, $locale);
            $this->translator->getCache()->removeItem($cacheId);

            $this->flashMessenger()
                ->addSuccessMessage('admin_i18n_translation_write_success');
        } catch (\Exception $e) {
            $this->flashMessenger()
                ->addErrorMessage('admin_i18n_translation_write_fail');
            if ($e->getMessage()) {
                $this->flashMessenger()
                    ->addErrorMessage($e->getMessage());
            }
        }

        if ($this->inLightbox() && empty($this->getFollowupUrl())) {
            return $this->getRefreshResponse();
        }

        return $this->forward()->dispatch(self::class, ['action' => 'home']);
    }

    /**
     * @param $object
     * @param $property
     *
     * @return mixed
     */
    protected function getPropertyValue($object, $property)
    {
        $reflectionProperty = $this->getProperty($object, $property);
        return $reflectionProperty->getValue($object);
    }

    /**
     * @param $object
     * @param $propertyName
     *
     * @return \ReflectionProperty
     */
    protected function getProperty($object, $propertyName): \ReflectionProperty
    {
        $key = get_class($object) . ":$propertyName";
        if (!isset($this->properties[$key])) {
            try {
                $property = new \ReflectionProperty($object, $propertyName);
                $property->setAccessible(true);
                $this->properties[$key] = $property;
            } catch (\ReflectionException $e) {
                throw new \RuntimeException($e);
            }
        }

        return $this->properties[$key];
    }

    /**
     * Check if locale matches existing locales
     *
     * @param string $locale local file name
     *
     * @return bool
     */
    protected function isValidLocal(string $locale): bool
    {
        $isValidLocal = false;
        foreach ($this->languages as $language) {
            if ($language["locale"] == $locale) {
                $isValidLocal = true;
                break;
            }
        }
        return $isValidLocal;
    }

    /**
     * Are we currently in a lightbox context?
     *
     * @return bool
     */
    protected function inLightbox()
    {
        return $this->getRequest()->getQuery('layout', 'no') === 'lightbox'
            || 'layout/lightbox' == $this->layout()->getTemplate();
    }

    /**
     * Construct an HTTP 205 (refresh) response. Useful for reporting success
     * in the lightbox without actually rendering content.
     *
     * @return \Laminas\Http\Response
     */
    protected function getRefreshResponse()
    {
        $response = $this->getResponse();
        $response->setStatusCode(205);
        return $response;
    }
}
