<?php
/**
 * Amsl API Controller
 *
 * PHP version 5
 *
 * Copyright (C) Leipzig University Library 2018.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category Finc
 * @package  Controller
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @author   Ulf Seltmann <seltmann@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Site
 */
namespace finc\Controller;

use VuFind\Controller\AbstractBase;
use Laminas\ServiceManager\ServiceLocatorInterface;

/**
 * Controller for the user account area.
 *
 * @category Finc
 * @package  Controller
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @author   Ulf Seltmann <seltmann@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Site
 */
class AmslResourceController extends AbstractBase
{
    const DEFAULT_TTL = 86400;

    /**
     * Amsl.ini configuration.
     *
     * @var \Laminas\Config\Config
     */
    protected $config = [];

    /**
     * @var mixed|null AMSL API configuration
     */
    protected $apiConfig;

    /**
     * HTTP client
     *
     * @var \Laminas\Http\Client
     */
    protected $httpClient;

    /**
     * Cache manager
     *
     * @var \VuFind\Cache\Manager
     */
    protected $cacheManager;

    /**
     * base url
     *
     * @var string
     */
    protected $baseUrl;

    /**
     * mapping of resource labels to be re-written for the query string
     *
     * @var array | null
     */
    protected $resourceMapping;

    /**
     * Constructor
     *
     * @param ServiceLocatorInterface $sm
     * @param \Laminas\Config\Config     $config     VuFind configuration
     * @param \VuFindHttp\HttpService $httpClient HttpClient
     * @param \VuFind\Cache\Manager $cacheManager Cache manager (optional)
     *
     */
    public function __construct(
        ServiceLocatorInterface $sm,
        \Laminas\Config\Config $config,
        \VuFindHttp\HttpService $httpClient,
        \VuFind\Cache\Manager $cacheManager = null,
        array $resourceMapping = null
    ) {
        parent::__construct($sm);
        $this->config = $config;
        $this->apiConfig = $config->API;
        $this->httpClient = $httpClient;
        $this->cacheManager = $cacheManager;
        $this->resourceMapping = $resourceMapping;
    }

    /**
     * Amsl action - controller method
     *
     * @return \Laminas\View\Model\ViewModel
     * @throws \Exception
     */
    public function homeAction()
    {
        // Make view
        $view = $this->createViewModel();
        $view->setTemplate('amsl/sources-list');
        $cacheKey = "rendered-{$this->layout()->userLang}";

        if ($data = $this->getCacheData($cacheKey)) {
            $view->rendered_html = $data;
        } else {
            try {
                $view->sources = $this->getData();
            } catch (\Exception $e) {
                $this->flashMessenger()->addMessage(
                    'resources_cannot_received',
                    'error'
                );
            }
            $rendered = $this->getViewRenderer()->render($view);
            $this->setCacheData($rendered, $cacheKey);
            $view->rendered_html = $rendered;
        }

        return $view;
    }

    /**
     * Retrieve amsl sources list, prepare and cache it
     *
     * @return array
     */
    protected function getData()
    {
        if (null == ($result = $this->httpClient->get($this->apiConfig->url))) {
            throw new \Exception(
                'Unexpected value: No api result received'
            );
        }
        if ($result->isSuccess()) {
            switch ($this->apiConfig->response_type) {
                case 'application/json':
                    $amsl_sources = json_decode($result->getBody(), true);
                    break;
                default:
                    throw new \Exception(
                        'Invalid argument: No valid header scheme defined'
                    );
                    break;
            }

            if (isset($amsl_sources)) {
                $results = $this->createSourceHierarchy($amsl_sources);
            }
        }

        return $results;
    }

    /**
     * Sorts the input array according to the values for main_key and sub_key
     *
     * @param array $amsl_sources
     *
     * @return array $out
     */
    protected function createSourceHierarchy(array $amsl_sources)
    {
        $mapping = $this->config->get('Mapping');
        $sources = [];

        foreach ($amsl_sources as $source) {
            $this->prepareSourceForView($source);
            if (isset($source[$mapping->main_key])) {
                if (isset($source[$mapping->sub_key])) {
                    $label = $this->renderLabel($mapping->sub_label, $source);
                    $sources[$source[$mapping->main_key]][$label] = $source;
                } else {
                    $sources[$source[$mapping->main_key]][$mapping->default_sub_label] = $source;
                }
            } else {
                if (isset($source[$mapping->sub_key])) {
                    $label = $this->renderLabel($mapping->sub_label, $source);
                    $default[$label] = $source;
                } else {
                    $default[$mapping->default_sub_label] = $source;
                }
            }
        }
        ksort($sources);
        $out = [];
        foreach ($sources as $main) {
            $label = $this->renderLabel($mapping->main_label, current($main));
            $out[$label] = $main;
        }
        if (isset($default)) {
            $out[$mapping->default_main_label] = $default;
        }

        if (!empty($mapping->sortBySourceLabel)) {
            uksort($out,'strcasecmp');
        }

        return $out;
    }

    /**
     * Helper funtion to render label
     *
     * @param string $pattern     Pattern
     * @param array  $input_array Input
     *
     * @return mixed
     */
    protected function renderLabel($pattern, $input_array)
    {
        $struct = [];
        $replace = [];
        preg_match_all('/\%\%(\w+)\%\%/', $pattern, $struct);
        foreach ($struct[1] as $key) {
            $replace[] = (isset($input_array[$key])) ? $input_array[$key] : '';
        }
        return str_replace(
            $struct[0],
            $replace,
            $pattern
        );
    }

    /**
     * Helper function to generate link and description
     *
     * @param array &$source Input
     *
     * @return void
     */
    protected function prepareSourceForView(&$source)
    {
        $mapping = $this->config->get('Mapping');

        $showDescription = $mapping->show_description ?? false;
        switch($showDescription) {
        case 1:
        case 'true':
        case 'api':
            if (isset($mapping->sub_description_key) && isset($source[$mapping->sub_description_key])) {
                $source["desc"] = $source[$mapping->sub_description_key];
                unset($source[$mapping->sub_description_key]);
            }
            break;
        case 'local':
            $source['desc'] = $this->translate('AMSLScopeNotes::'.$source[$mapping->sub_key], [], '');
            break;
        case 0:
        case 'false':
        default:
            // do nothing
        }

        if(!empty($mapping->show_link) && !empty($source[$mapping->sub_key])) {

            $searchTerm = $source[$mapping->sub_key];
            // map the source label to the correct SOLR search term
            $searchTerm = $this->resourceMapping[$searchTerm] ?? $searchTerm;

            if (!$this->baseUrl) {
                $urlHelper = $this->getViewRenderer()->plugin('url');
                $this->baseUrl = $urlHelper('search-results') . '?filter%5B%5D=mega_collection%3A%22';
            }

            if (!empty($searchTerm)) {
                $source["href"] = $this->baseUrl . urlencode($searchTerm) . '%22';
            }
        }
    }

    /**
     * @return string | null
     * @throws \Exception
     */
    private function getCacheData($tag = ''): ?string
    {
        $cache = $this->cacheManager->getCache('object', $this->getCacheNamespace());

        $cacheKey = $this->getCacheKey($tag);
        $ttl = $this->apiConfig['ttl'] ?? self::DEFAULT_TTL;

        if($cache) {
            $cache->getOptions()->setTtl($ttl);
            $cache->clearExpired();
        }

        return $cache->getItem($cacheKey);
    }

    /**
     * Helper function for writing data to controller specific cache
     *
     * @param  string $value Cache value to write
     * @param  string $tag   optional tag to be used for cache key
     * @throws \Exception
     */
    private function setCacheData($value, $tag = '')
    {
        $cache = $this->cacheManager->getCache('object', $this->getCacheNamespace());
        $cacheKey = $this->getCacheKey($tag);
        $cache->setItem($cacheKey, $value);
    }

    /**
     * Get the namespace to use for caching amls responses.
     *
     * @return string
     */
    protected function getCacheNamespace()
    {
        return 'amsl';
    }

    /**
     * Helper function to generate config-specific cache key
     *
     * @param  string $tag optional tag to make cache more specific
     * @return string
     */
    protected function getCacheKey($tag = '')
    {
        return md5(json_encode($this->config->toArray()).$tag);
    }
}
