<?php
/**
 * MyResearch Controller
 *
 * PHP version 5
 *
 * Copyright (C) Villanova University 2010.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  Controller
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Site
 */
namespace finc\Controller;

use VuFind\Exception\Forbidden as ForbiddenException;
use VuFind\Exception\ListPermission as ListPermissionException;
use VuFind\Search\RecommendListener;
use Laminas\Log\LoggerAwareInterface as LoggerAwareInterface;
use Laminas\Mvc\MvcEvent as MvcEvent;

/**
 * Controller for the user account area.
 *
 * @category VuFind
 * @package  Controller
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Site
 */
class MyResearchController extends \VuFind\Controller\MyResearchController implements
    LoggerAwareInterface
{
    use \VuFind\Log\LoggerAwareTrait;
    use CustomTraits\AcquisitionTrait;
    use CustomTraits\EmailProfileTrait;
    use CustomTraits\ResetPasswordTrait;

    /**
     * Execute the request
     *
     * @param \Laminas\Mvc\MvcEvent $event Event
     *
     * @return mixed
     * @throws Exception\DomainException
     */
    public function onDispatch(MvcEvent $event)
    {
        if ($redirect = $this->getRequest()->getQuery()->get('redirect')) {
            $redirect = urldecode($redirect);
            //if we have a "redirect" param, the user should be lead to the given url,
            //at least as long as his login is successful
            //we achieve that by forcing the follow-up url to be the redirect
            //but watch out, the redirect via followup will only work if the redirect url
            //is schemed correctly
            if (!is_string($redirect)) {
                $this->flashMessenger()->addErrorMessage('Invalid data type for redirect');
            } elseif (!preg_match('/^https?\:\/\//', $redirect)) {
                $this->flashMessenger()->addErrorMessage('Invalid Redirect: ' . $redirect);
            } else {
                $this->followup()->store(['finc-redirect' => $redirect]);
            }
        }
        return parent::onDispatch($event);
    }

    /**
     * Send user's saved favorites from a particular list to the view
     *
     * @return mixed
     */
    public function mylistAction()
    {
        // Fail if lists are disabled:
        if (!$this->listsEnabled()) {
            throw new ForbiddenException('Lists disabled');
        }

        // Check for "delete item" request; parameter may be in GET or POST depending
        // on calling context.
        $deleteId = $this->params()->fromPost(
            'delete', $this->params()->fromQuery('delete')
        );
        if ($deleteId) {
            $deleteSource = $this->params()->fromPost(
                'source',
                $this->params()->fromQuery('source', DEFAULT_SEARCH_BACKEND)
            );
            // If the user already confirmed the operation, perform the delete now;
            // otherwise prompt for confirmation:
            $confirm = $this->params()->fromPost(
                'confirm', $this->params()->fromQuery('confirm')
            );
            if ($confirm) {
                $layout = $this->params()->fromPost(
                    'layout', $this->params()->fromQuery('layout')
                );

                /* #17712 not necessary to fetch items after deleting by ajax */
                $success = $this->performDeleteFavorite($deleteId, $deleteSource);
                if ($layout === 'lightbox' || $success !== true) {
                    return $success;
                }

            } else {
                return $this->confirmDeleteFavorite($deleteId, $deleteSource);
            }
        }

        // If we got this far, we just need to display the favorites:
        try {
            $runner = $this->serviceLocator->get('VuFind\Search\SearchRunner');

            // We want to merge together GET, POST and route parameters to
            // initialize our search object:
            $request = $this->getRequest()->getQuery()->toArray()
                + $this->getRequest()->getPost()->toArray()
                + ['id' => $this->params()->fromRoute('id')];

            // Set up listener for recommendations:
            $rManager = $this->serviceLocator
                ->get('VuFind\Recommend\PluginManager');
            $setupCallback = function ($runner, $params, $searchId) use ($rManager) {
                $listener = new RecommendListener($rManager, $searchId);
                $listener->setConfig(
                    $params->getOptions()->getRecommendationSettings()
                );
                $listener->attach($runner->getEventManager()->getSharedManager());
            };

            $results = $runner->run($request, 'Favorites', $setupCallback);
            return $this->createViewModel(
                ['params' => $results->getParams(), 'results' => $results, 'context' => "favorite"]
            );
        } catch (ListPermissionException $e) {

            /* #14751 disallow private lists for crawlers (for finc #17318) - RL */
            if ($headMeta = $this->getViewRenderer()->plugin('headmeta')) {
                $headMeta()->prependName('robots', 'noindex');
            }
            /* #14751 disallow private lists for crawlers (for finc #17318) - END */

            if (!$this->getUser()) {
                return $this->forceLogin();
            }
            throw $e;
        }
    }

    /**
     * Delete record
     * Overrides Vufind method without flash messages
     *
     * @param string $id     ID of record to delete
     * @param string $source Source of record to delete
     *
     * @return bool|mixed
     *
     * @throws \Exception
     **/
    public function performDeleteFavorite($id, $source)
    {
        // Force login:
        $user = $this->getUser();
        if (!$user) {
            return $this->forceLogin();
        }

        // Load/check incoming parameters:
        $listID = $this->params()->fromRoute('id');
        $listID = empty($listID) ? null : $listID;
        if (empty($id)) {
            throw new \Exception('Cannot delete empty ID!');
        }

        // Perform delete #17712 without setting messages
        if (null !== $listID) {
            // ...Specific List
            $table = $this->getTable('UserList');
            $list = $table->getExisting($listID);
            $list->removeResourcesById($user, [$id], $source);
        } else {
            // ...My Favorites
            $user->removeResourcesById([$id], $source);
        }

        // All done -- return true to indicate success.
        return true;
    }
}
