<?php
/**
 * Hierarchy Tree Renderer for the JS_Tree plugin
 *
 * PHP version 5
 *
 * Copyright (C) Villanova University 2010.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @category VuFind
 * @package  HierarchyTree_Renderer
 * @author   Luke O'Sullivan <l.osullivan@swansea.ac.uk>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:plugins:hierarchy_components Wiki
 */
namespace finc\Hierarchy\TreeRenderer;

use Laminas\Config\Config;

/**
 * Hierarchy Tree Renderer
 *
 * This is a helper class for producing hierarchy trees.
 *
 * @category VuFind
 * @package  HierarchyTree_Renderer
 * @author   Luke O'Sullivan <l.osullivan@swansea.ac.uk>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:plugins:hierarchy_components Wiki
 */
class JSTree extends \VuFind\Hierarchy\TreeRenderer\JSTree
{
    /**
     * @var Config treeRendererConfig
     */
    protected $treeRendererConfig;

    public function setConfig(Config $treeRendererConfig) {
        $this->treeRendererConfig = $treeRendererConfig;
    }

    /**
     * Use the router to build the appropriate URL based on context
     *
     * @param object $node    JSON object of a node/top node
     * @param string $context Record or Collection
     *
     * @return string
     */
    protected function getContextualUrl($node, $context)
    {
        //set default node type to record
        $type = 'record';
        if ($node->type == 'collection') {
            $type = 'collection';
        } elseif (!empty($node->children)) {
            //this is the finc-specific node type for
            //hierarchy parents
            $type = 'parent';
        }
        return $this->getUrlFromRouteCache($type, $node->id);
    }

    /**
     * Get the URL for a record and cache it to avoid the relatively slow routing
     * calls.
     *
     * @param string $route Route
     * @param string $id    Record ID
     *
     * @return string URL
     */
    protected function getUrlFromRouteCache($route, $id)
    {
        static $cache = [];
        if (!isset($cache[$route])) {
            if ($route == 'collection') {
                $tab = $this->treeRendererConfig->linkedTabForCollections ?? 'HierarchyTree';
                $params = [
                    'id' => '__record_id__',
                    'tab' => $tab
                ];
                if ($tab === 'HierarchyTree') {
                    $options = [
                        'query' => [
                            'recordID' => '__record_id__'
                        ]
                    ];
                }
                $cache[$route] = $this->router->fromRoute($route, $params, $options ?? []);
            } elseif ($route == 'parent') {
                // this is nearly the same as for collections
                // but links on the record view i.e. record route
                $tab = $this->treeRendererConfig->linkedTabForParents;
                if (empty($tab)) $tab = 'HierarchyTree';
                $params = [
                    'id' => '__record_id__',
                    'tab' => $tab
                ];
                if ($tab === 'HierarchyTree') {
                    $options = [
                        'query' => [
                            'recordID' => '__record_id__'
                        ]
                    ];
                }
                // use route 'record' here to avoid linking to the
                // collection view
                $cache[$route] = $this->router->fromRoute('record', $params, $options ?? []);
            } else {
                $tab = $this->treeRendererConfig->linkedTabDefault ?? 'Description';
                $params = [
                    'id' => '__record_id__',
                    'tab' => $tab
                ];
                $cache[$route] = $this->router->fromRoute($route, $params);
            }
        }
        return str_replace('__record_id__', $id, $cache[$route]);
    }

    /**
     * @return \VuFind\Hierarchy\TreeDataSource\AbstractBase
     */
    public function getDataSource()
    {
        if (!isset($this->dataSource)) {
            $this->dataSource = parent::getDataSource();
        }
        $this->dataSource->setRecordDriver($this->recordDriver);
        return $this->dataSource;
    }

    /**
     * Almost complete copy from vufind's JStree:
     * remove just title attribute
     *
     * @param object $node        JSON object of a node/top node
     * @param string $context     Record or Collection
     * @param string $hierarchyID Collection ID
     *
     * @return array
     *
     * @inheritdoc
     */
    protected function buildNodeArray($node, $context, $hierarchyID)
    {
        $escaper = new \Laminas\Escaper\Escaper('utf-8');
        $ret = [
            'id' => preg_replace('/\W/', '-', $node->id),
            'text' => $escaper->escapeHtml($node->title),
            'li_attr' => [
                'recordid' => $node->id
            ],
            'a_attr' => [
                'href' => $this->getContextualUrl($node, $context),
                // Remove redundant title attribute - refs #18621
                //'title' => $node->title
            ],
            'type' => $node->type
        ];
        if (isset($node->children)) {
            $ret['children'] = [];
            for ($i = 0;$i < count($node->children);$i++) {
                $ret['children'][$i] = $this
                    ->buildNodeArray($node->children[$i], $context, $hierarchyID);
            }
        }
        return $ret;
    }
}
