<?php
/**
 * Catalog Connection Class
 *
 * This wrapper works with a driver class to pass information from the ILS to
 * VuFind.
 *
 * PHP version 5
 *
 * Copyright (C) Villanova University 2007.
 * Copyright (C) Leipzig University Library 2016.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  ILS_Drivers
 * @author   Andrew S. Nagy <vufind-tech@lists.sourceforge.net>
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @author   André Lahmann <lahmann@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:plugins:ils_drivers Wiki
 */
namespace finc\ILS;

use VuFind\I18n\Translator\TranslatorAwareInterface;

/**
 * Catalog Connection Class
 *
 * This wrapper works with a driver class to pass information from the ILS to
 * VuFind.
 *
 * @category VuFind
 * @package  ILS_Drivers
 * @author   Andrew S. Nagy <vufind-tech@lists.sourceforge.net>
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @author   André Lahmann <lahmann@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:plugins:ils_drivers Wiki
 */
class Connection extends \VuFind\ILS\Connection implements TranslatorAwareInterface
{
    /**
     * Check Email Hold
     *
     * A support method for checkFunction(). This is responsible for checking
     * the driver configuration to determine if the system supports storage
     * retrieval requests.
     *
     * @param array $functionConfig The email hold configuration values
     * @param array $params         An array of function-specific params (or null)
     *
     * @return mixed On success, an associative array with specific function keys
     * and values either for placing requests via a form; on failure, false.
     */
    protected function checkMethodEmailHold($functionConfig, $params)
    {
        $response = false;

        // $params doesn't include all of the keys used by
        // placeStorageRetrievalRequest, but it is the best we can do in the context.
        /*$check = $this->checkCapability(
            'placeEmailHold', [$params ?: []]
        );*/
        if (isset($functionConfig['HMACKeys'])) {
            $response['HMACKeys'] = explode(':', $functionConfig['HMACKeys']);
            if (isset($functionConfig['extraFields'])) {
                $response['extraFields'] = $functionConfig['extraFields'];
            }
            if (isset($functionConfig['emailProfileSelector'])) {
                $response['emailProfileSelector'] = $functionConfig['emailProfileSelector'];
            }
            if (isset($functionConfig['helpText'])) {
                $response['helpText'] = $this->getHelpText(
                    $functionConfig['helpText']
                );
            }
        }
        return $response;
    }

    /**
     * Check ILLRequests
     *
     * A support method for checkFunction(). This is responsible for checking
     * the driver configuration to determine if the system supports ILL requests.
     *
     * @param array $functionConfig The ILL request configuration values
     * @param array $params         An array of function-specific params (or null)
     *
     * @return mixed On success, an associative array with specific function keys
     * and values either for placing requests via a form; on failure, false.
     */
    protected function checkMethodILLRequests($functionConfig, $params)
    {
        if (
            method_exists($this, 'getMyILLRequests')
            &&
            isset($functionConfig['daiaILLpattern'])
        ) {
            return ['daiaILLpattern' => $functionConfig['daiaILLpattern']];
        }

        return false;
    }

    /**
     * Get access to the driver object.
     * Extends parent function with a bypass of failover on InitException
     *
     * @param bool $init Should we initialize the driver (if necessary), or load it
     * "as-is"?
     *
     * @throws \Exception
     * @throws InitException
     * @return object
     */
    public function getDriver($init = true)
    {
        if (null === $this->driver) {
            $this->setDriver($this->driverManager->get($this->config->driver));
        }
        if (!$this->driverInitialized && $init) {
            try {
                $this->initializeDriver();
            } catch (InitException $e) {
                throw $e;
            } catch (\Exception $e) {
                if (!$this->failOverToNoILS()) {
                    throw $e;
                }
            }
        }
        return $this->driver;
    }

    /**
     * Get Hidden Login Mode
     *
     * This is responsible for indicating whether login should be hidden.
     *
     * Extends parent function with a bypass of failover on InitException
     *
     * @throws \Exception
     * @throws InitException
     * @return bool true if the login should be hidden, false if not
     */
    public function loginIsHidden()
    {
        // Graceful degradation -- return false if no method supported.
        try {
            return $this->checkCapability('loginIsHidden')
                ? $this->getDriver()->loginIsHidden() : false;
        } catch (InitException $e) {
            throw $e;
        } catch (\Exception $e) {
            if ($this->failOverToNoILS()) {
                return call_user_func_array([$this, __METHOD__], func_get_args());
            }
            throw $e;
        }
    }

    /**
     * Get Offline Mode
     *
     * This is responsible for returning the offline mode
     *
     * @param bool $healthCheck Perform a health check in addition to consulting
     * the ILS status?
     *
     * @return string|bool "ils-offline" for systems where the main ILS is offline,
     * "ils-none" for systems which do not use an ILS, false for online systems.
     *
     * @deprecated Backport of VuFind 8.1 (healthCheckId) - remove after upgrade
     */
    public function getOfflineMode($healthCheck = false)
    {
        // If we have NoILS failover configured, force driver initialization so
        // we can know we are checking the offline mode against the correct driver.
        if ($this->hasNoILSFailover()) {
            $this->getDriver();
        }
        $hasOfflineMode = $this->checkCapability('getOfflineMode');
        // If we need to perform a health check, try to do a random item lookup
        // before proceeding.
        if ($healthCheck && $this->config->healthCheckId) {
            if (empty($this->getStatus($this->config->healthCheckId)) && $hasOfflineMode) {
                // show different return string to separate between basic availability and PAIA later?
                return 'ils-offline';
            }
        }

        // If we're encountering failures, let's go into ils-offline mode if
        // the ILS driver does not natively support getOfflineMode().
        $default = $this->failing ? 'ils-offline' : false;

        // Graceful degradation -- return false if no method supported.
        return $hasOfflineMode ? $this->getDriver()->getOfflineMode() : $default;
    }
}
