<?php
/**
 * finc specific model for Solr records based on the stock
 * VuFind\RecordDriver\SolrDefault
 *
 * PHP version 5
 *
 * Copyright (C) Leipzig University Library 2015.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  RecordDrivers
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     http://vufind.org/wiki/vufind2:record_drivers Wiki
 */
namespace finc\RecordDriver;

use VuFind\Db\Row\Resource;

/**
 * Bundles serialization functionality used to store records in resource table
 *
 * @category VuFind
 * @package  RecordDrivers
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     http://vufind.org/wiki/vufind2:record_drivers Wiki
 */
trait FincRecordSerializationTrait
{
    /**
     * @var array $serializationMapping List of functions whose return values
     * are serialized to the resource table
     */
    protected $serializationMapping = [
        'title' => 'getTitle',
        'shortTitle' => 'getShortTitle',
        'publishPlace' => 'getPlacesOfPublication',
        'publisher' => 'getPublishers',
        'isbn' => 'getISBNs',
        'issn' => 'getISSNs',
        'DOI' => 'getCleanDOI',
        'description' => 'getSummary',
        'callnumber-raw' => 'getCallNumbers',
        'container_title' => 'getContainerTitle',
        'format' => 'getFormats',
        'author' => 'getPrimaryAuthors',
        'author2' => 'getSecondaryAuthors',
        'additional_authors' => 'getAdditionalAuthors',
    ];

    /**
     * @var array $savedInSeperateFields List of fields to be store in the
     * respective DB field. Schema 'local_field_name' => 'db_field_name'
     */
    protected $savedInSeperateFields = [
        'publishDate' => 'year',
        'author' => 'author',
        'id' => 'record_id'
    ];

    /**
     * Serializes values from this record's field property to a json string that
     * will be used as a substitute in the title field of VuFind's resource table
     *
     * @param  bool $cropped true if only use single values from arrays shall be used
     * @return false|string
     * @throws FincRecordSerializationException
     */
    public function serializeForResourceTable()
    {
        $output = new \stdClass();
        foreach ($this->serializationMapping as $key => $function) {
            $retval = $this->tryMethod($function);
            if (!empty($retval)) {
                $output->$key = $retval;
            }
        }
        return $output;
    }

    /**
     * Populates the fields property of the attached record
     * with values from the resource table
     *
     * @param Resource $input row from the resource table,
     * containing a title as serialized by
     * \finc\RecordDriver\RecordSerializationTrait::serializeForResourceTable
     */
    public function deserializeFromResourceTable(Resource $input)
    {
        foreach ($this->savedInSeperateFields as $local_key => $db_key) {
            // populate fields from predefined DB table columns
            if (isset($input->$db_key)) {
                $this->fields[$local_key] = $input->$db_key;
            }
        }
        if (isset($input->title)) {
            $this->fields['title'] = ucfirst($input->title);
        }
        // populate others from serialized info in title field
        $input = json_decode($input->extra_metadata, true);
        // the json input is an associative array keyed with
        // the exported field names, thus we can array_intersect
        // the input here
        $this->fields = array_merge(
            $this->fields,
            // beware the order here, we need $input's field values
            array_intersect_key(
                $input,
                $this->serializationMapping
            )
        );
    }

    /**
     * @return false|string
     * @throws FincRecordSerializationException
     */
    public function getExtraResourceMetadata()
    {
        return $this->serializeForResourceTable();
    }
}

/**
 * thrown on serialization errors from this trait
 * Class SerializationException
 * @package finc\RecordDriver
 */
class FincRecordSerializationException extends \Exception
{
}
