<?php
/**
 * Munger Injection Factory
 *
 * A Delegator Factory that registers several listeners at events triggered by the
 * VuFind\Search service.
 *
 * PHP version 7
 *
 * Copyright (C) Leipzig University Library 2019.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  Finc/Service
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Page
 */
namespace finc\Service;

use Interop\Container\ContainerInterface;
use VuFindSearch\Query\Query;
use VuFindSearch\Query\QueryGroup;
use VuFindSearch\Service as SearchService;
use Laminas\Config\Config;
use Laminas\EventManager\EventInterface;
use Laminas\ServiceManager\Factory\DelegatorFactoryInterface;

/**
 * Munger Injection Factory
 *
 *
 * @category VuFind
 * @package  Finc/Service
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Page
 */
class MungerInjectionDelegatorFactory implements DelegatorFactoryInterface
{
    /**
     * @var SearchService
     */
    protected $instance;

    /**
     * @var array configuration for pre-search munging
     */
    protected $preMungerConfig;

    /**
     * @var array search types to be excluded from lucene query syntax detection
     */
    protected $override_advanced_lucene_detection;

    /**
     * Creates a delegator of VuFind/Search to register several listeners.
     *
     * @param ContainerInterface $container
     * @param string $name
     * @param callable $callback
     * @param array|null $options
     *
     * @return mixed
     */
    public function __invoke(
        ContainerInterface $container,
        $name,
        callable $callback,
        array $options = null
    ) {
        $instance = call_user_func($callback);

        $searchConfig = $container->get('VuFind\Config')->get('searches');
        $e = $instance->getEventManager()->getSharedManager();

        if ($this->validateMungerConfig($searchConfig)) {
            $e->attach(
                'VuFindSearch',
                'pre',
                function (EventInterface $event) {
                    $params = $event->getParams();
                    if (isset($params['query'])) {
                        $params['query'] = $this->preMunge($params['query']);
                    }
                }
            );
        }
        if (isset($searchConfig->General->override_advanced_lucene_detection)) {
            $this->override_advanced_lucene_detection = array_flip($searchConfig->General->override_advanced_lucene_detection->toArray());
            if (!empty($this->override_advanced_lucene_detection)) {
                $e->attach('VuFindSearch', 'pre', [$this, 'overrideLuceneSyntaxHelper']);
            }
        }
        return $instance;
    }

    /**
     * Escapes colons in Queries or recursively in QueryGroups.
     * This prevents queries from being interpreted as advanced queries in Lucene syntax.
     * cf. \VuFindSearch\Backend\Solr\LuceneSyntaxHelper::containsAdvancedLuceneSyntax
     *
     * @param Query|QueryGroup $queryOrGroup
     *
     * @return mixed
     */
    private function preMunge($queryOrGroup)
    {
        if ($queryOrGroup instanceof QueryGroup) {
            $handler = $queryOrGroup->getReducedHandler();
            if (is_null($handler) || isset($this->preMungerConfig[$handler])) {
                foreach ($queryOrGroup->getQueries() as $query) {
                    $this->preMunge($query);
                }
            }
        } elseif ($mungerConfig = $this->preMungerConfig[$queryOrGroup->getHandler()] ?? null) {
            $queryOrGroup->setString(
                // apply preg_replace as provided by config
                preg_replace($mungerConfig['pattern'], $mungerConfig['replace'], $queryOrGroup->getString())
            );
        }
        return $queryOrGroup;
    }

    /**
     * @param Config $searchConfig
     */
    protected function validateMungerConfig($searchConfig) {

        $config = $searchConfig->PreMunge;
        if (empty($config)) {
            // an empty configuration is always valid
            return null;
        }
        $preMungerConfig = $config->toArray();
        foreach ($preMungerConfig as $handler => $handlerConfig) {
            if (!isset($handlerConfig['pattern'])
                ||
                !isset($handlerConfig['replace'])
            ) {
                throw new \ConfigurationException("PreMunge configuration for $handler is invalid");
            }
            try {
                // try preg_replace so the regex engine tells us about more errors
                preg_replace($handlerConfig['pattern'], $handlerConfig['replace'], '');
            } catch (\ErrorException $e) {
                throw new \ConfigurationException("PreMunge configuration for $handler is invalid. Regex error: ".'"'.$e->getMessage().'"');
            }
        }
        $this->preMungerConfig = $preMungerConfig;
        return true;
    }

    public function overrideLuceneSyntaxHelper(EventInterface $event)
    {
        $params = $event->getParams();
        if (
            $params['context'] == 'search'
            && isset($params['query'])
            && $this->isLuceneReducible($params['query'])
        ) {
            $builder = $event->getTarget()->getQueryBuilder();
            $currentHelper = $builder->getLuceneHelper();
            $newHelper = new LuceneSyntaxHelperOverride($currentHelper);
            $builder->setLuceneHelper($newHelper);
        }
    }


    /**
     * Escapes colons in Queries or recursively in QueryGroups.
     * This prevents queries from being interpreted as advanced queries in Lucene syntax.
     * cf. \VuFindSearch\Backend\Solr\LuceneSyntaxHelper::containsAdvancedLuceneSyntax
     *
     * @param Query|QueryGroup $queryOrGroup
     *
     * @return mixed
     */
    private function isLuceneReducible($queryOrGroup)
    {
        if ($queryOrGroup instanceof QueryGroup) {
            $handler = $queryOrGroup->getReducedHandler();
            if (is_null($handler) || isset($this->override_advanced_lucene_detection[$handler])) {
                foreach ($queryOrGroup->getQueries() as $query) {
                    // cycle through all recursive subqueries, if any of these matches
                    // the whole query should not be checked for lucene syntax
                    if ($this->isLuceneReducible($query)) {
                        return true;
                    }
                }
            }
        } elseif (isset($this->override_advanced_lucene_detection[$queryOrGroup->getHandler()])) {
            return true;
        }
        return false;
    }
}
