<?php
/**
 * External link view helper
 *
 * PHP version 7
 *
 * Copyright (C) Leipzig University Library 2020.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  View_Helpers
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development Wiki
 */
namespace finc\View\Helper\Root;

use Laminas\View\Helper\EscapeHtml;
use Laminas\View\Helper\EscapeHtmlAttr;

/**
 * External link view helper
 *
 * @category VuFind
 * @package  View_Helpers
 * @author   Dorian Merz <merz@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development Wiki
 */
class ExternalLink extends \Laminas\View\Helper\AbstractHelper
{
    /**
     * Default html attributes for external links
     *
     * @var string[]
     */
    protected $defaultAttributes = [
        'target' => "_blank",
        'rel' => 'noopener'
    ];

    /**
     * Escape manuelly flag
     *
     * @var bool
     */
    protected $manualEscape;

    /**
     * Renders an anchor element (hyperlink) to an external website
     *
     * @param string $href           URL of the external website
     * @param string $label          desired label or translation token
     * @param array  $attributes     more attributes to apply to the a element,
     *                               key-value-pairs
     * @param bool   $manualEscape   true if $label conatins html to display and
     *                               $attributes are escped before calling method
     *
     * @return string
     */
    public function __invoke(
        $href,
        $label = '',
        $attributes = [],
        $manualEscape = false
    ) {
        $this->manualEscape = $manualEscape;

        $link = '<a href="' . $href . '"';
        $attributes = array_merge($this->defaultAttributes, $attributes);
        foreach ($attributes as $key => $value) {
            $link .= ' ' . $key . '="' . $this->escapeHtmlAttr($value) . '"';
        }
        $link .= '>';
        if (empty($label)) {
            $label = $href;
        }
        $link .= $this->escapeHtml($label) . '</a>';

        return $link;
    }

    /**
     * Helper function to provide access to EscapeHTML ViewHelper
     * and esacpe lable if not disabled
     *
     * @param string $html html to escape
     *
     * @return string
     */
    protected function escapeHtml($html)
    {
        if ($this->manualEscape) {
            return $html;
        }
        $htmlEscaper = $this->getView()->plugin('escapeHtml');
        return $htmlEscaper($html);
    }

    /**
     * Helper function to provide access to EscaperHTML Attribute ViewHelper
     * and esacpe HTML attributes if not disabled
     *
     * @param string $label translation token
     *
     * @return string
     */
    protected function escapeHtmlAttr($htmlAttr)
    {
        if ($this->manualEscape) {
            return $htmlAttr;
        }
        $htmlAttrEscaper = $this->getView()->plugin('escapeHtmlAttr');
        return $htmlAttrEscaper($htmlAttr);
    }
}
