<?php
/**
 * Record driver view helper
 *
 * PHP version 7
 *
 * Copyright (C) Villanova University 2010.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  View_Helpers
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development Wiki
 */
namespace finc\View\Helper\Root;

use finc\Rewrite\EblRewrite;

/**
 * Record driver view helper
 *
 * @category VuFind
 * @package  View_Helpers
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development Wiki
 */
class Record extends \VuFind\View\Helper\Root\Record
{
    /**
     * VuFind configuration
     *
     * @var \Laminas\Config\Config
     */
    protected $config;

    /**
     * Authentication manager
     *
     * @var \VuFind\Auth\Manager
     */
    protected $manager;

    /**
     * URL helper
     *
     * @var \Laminas\View\Helper\Url
     */
    protected $url;

    /**
     * Rewriter
     *
     * @var EblRewrite $rewrite
     */
    protected $rewrite;

    /**
     * Resolver configuration
     *
     * @var \Laminas\Config\Config
     */
    protected $resolverConfig;

    /**
     * Mapping for facets on style based icons
     *
     * @var \Laminas\Config\Config
     */
    protected $iconConfig;

    /**
     * Type of Mapping for record on its icon
     *
     * @var string
     */
    protected $iconMappingType;

    /**
     * Mapping for style based record icons
     *
     * @var array
     */
    protected $iconMappings = [];

    /**
     * Handling for array of entry keys
     *
     * True  => concatenate keys for mapping
     * False => use first key only for mapping
     *
     * @var string
     */
    protected $concatKeysForIconMapping;

    /**
     * Base css class for displaying icons
     * Like 'fa' needed by font awesome
     *
     * @var string
     */
    protected $baseIconClass;

    /**
     * Constructor
     *
     * @param \Laminas\Config\Config   $config         VuFind configuration
     * @param \Laminas\View\Helper\Url $helper         URL helper
     * @param \VuFind\Auth\Manager     $manager        AuthManager
     * @param mixed                    $rewrite        Ebl rewrite service for VuFind
     * @param mixed                    $resolverConfig Resolver.ini
     * @param mixed                    $iconConfig     iconMapping.ini
     */
    public function __construct(
        $config,
        \Laminas\View\Helper\Url $helper,
        \VuFind\Auth\Manager $manager,
        $rewrite,
        $resolverConfig,
        $iconConfig = null
    ) {
        parent::__construct($config);
        $this->url = $helper;
        $this->manager = $manager;
        $this->rewrite = $rewrite;
        $this->resolverConfig = $resolverConfig;

        if ($iconConfig) {
            $this->iconConfig = $iconConfig;
            $this->iconMappingType = $this->iconConfig
                ->StyleBasedIcons->type ?? 'None';
            $this->iconMappings = isset(
                $this->iconConfig
                    ->StyleBasedIcons->{$this->iconMappingType}
            ) ? $this->iconConfig
                ->StyleBasedIcons->{$this->iconMappingType}->toArray() : [];
            $this->concatKeysForIconMapping = isset(
                $this->iconConfig
                    ->StyleBasedIcons->concatenateKeys
            ) && $this->iconConfig
                ->StyleBasedIcons->concatenateKeys ? true : false;
            $this->baseIconClass = $this->iconConfig->StyleBasedIcons
                    ->baseIconClass ?? 'fa';
        }
    }

    /**
     * Render a (list of) record icons.
     *
     * @param string $tpl         phtml for record icon, default: record-icon.phtml
     * @param bool   $setIconText Return additional access / format. Default: true
     *
     * @return string
     */
    public function getRecordIcon(
        string $tpl = 'record-icon',
        bool $setIconText = true
    ): string {
        $iconType = $this->getRecordIconType();
        $iconClass = $this->getRecordIconClass($iconType);
        return $this->renderTemplate(
            $tpl . '.phtml',
            [
                'iconClass' => $this->baseIconClass . ' ' . $iconClass,
                'setIconText' => $setIconText
            ]
        );
    }

    /**
     * Get the CSS class used to properly render an icon for given value
     *
     * @param string $key (identifier) Value to convert into CSS class. Default: null
     *
     * @return string
     */
    public function getRecordIconClass(string $key = "default"): string
    {
        return $this->iconMappings[$key] ??
          $this->iconMappings["default"] ?? "unknown";
    }

    /**
     * Get the CSS class used to properly render an icon for given value
     *
     * @param string|null $iconType  icon type
     * @param string|null $iconClass css class for icon
     * @param string      $classfile Define alternative file for icon text without
     *
     * @return string
     */
    public function getRecordIconText(
        string $iconType = null,
        string $iconClass = null,
        string $classfile = 'record-icon-text'
    ): string {
        if (null === $iconType) {
            $iconType = $this->getRecordIconType();
        }
        if (null === $iconClass) {
            $iconClass = $this->getRecordIconClass($iconType);
        }
        return $this->renderTemplate(
            $classfile . '.phtml',
            compact('iconType', 'iconClass')
        );
    }

    /**
     * Get icon by record type
     *
     * @return string
     */
    protected function getRecordIconType()
    {
        $recordType = $this->driver->getRecordType();
        switch ($recordType) {
        case 'lido':
            return "object";
        case 'missing':
        case 'marcfincpda':
            return $recordType;
        default:
            $types = explode('_', $this->iconMappingType);
            $recordType = '';
            foreach ($types as $type) {
                $recordTypes = $this->driver->tryMethod("get{$type}") ?? ['default'];
                $recordTypes = $this->concatKeysForIconMapping ?
                  implode('', $recordTypes) : $recordTypes[0] ?? 'default';
                $recordType .= strtolower(
                    str_replace([' ','-',','], '', $recordTypes)
                );
            }
            return $recordType;
        }
    }

    /**
     * Returns if style based icons should be shown (if covers are disabled!)
     *
     * @return bool
     */
    public function showStyleBasedIcons()
    {
        return $this->config->Content->showStyleBasedIcons ?? false;
    }

    /**
     * Render the link of the specified type.
     *
     * @param string $type    Link type
     * @param string $lookfor String to search for at link
     *
     * @return string
     */
    public function getLink($type, $lookfor)
    {
        $lookfor = (
            $type == 'author'
                ? $this->removeAuthorDates($lookfor) : $lookfor
        );
        /* remove invalid new line in link templates,
        remove after fixed in templates or VuFind parent */
        return str_replace(["\n"], '', parent::getLink($type, $lookfor) ?? '');
    }

    /**
     * Get all the links associated with this record.  Returns an array of
     * associative arrays each containing 'desc' and 'url' keys.
     *
     * @param bool $openUrlActive Is there an active OpenURL on the page?
     *
     * @return array
     */
    public function getLinkDetails($openUrlActive = false)
    {
        $links = parent::getLinkDetails($openUrlActive);
        return $this->rewriteLinks($links);
    }

    /**
     * Map list of multiple dates to a view of date range as string e.g.
     * startdate - enddate Sorting of array should be managed by backend.
     *
     * @param array $dates Dates
     *
     * @return mixed
     */
    public function mapDateListToRangeView($dates)
    {
        if (is_array($dates)) {
            if (count($dates) == 1) {
                return $dates[0];
            } else {
                return array_shift($dates) . '-' . array_pop($dates);
            }
        }
        return $dates;
    }

    /**
     * Remove author dates from author string (used for using author names as search
     * term).
     *
     * @param string $author Author data
     *
     * @return mixed
     */
    public function removeAuthorDates($author)
    {
        $match = [];
        $regex
            = '/^(\s|.*)\s(fl.\s|d.\s|ca.\s|\*)*\s?(\d{4})\??(\sor\s\d\d?)?\s?'
              . '(-|–)?\s?(ca.\s|after\s|†)?(\d{1,4})?(.|,)?$/Uu';
        if (preg_match($regex, $author, $match)) {
            $author = (isset($match[1])) ? trim($match[1]) : $author;
        }
        // delete unnormalized characters of gallica ressource with source_id:20
        if (preg_match('/(.*)(\d.*)/Uus', $author, $match)) {
            $author = (isset($match[1])) ? trim($match[1]) : $author;
        }
        return $author;
    }

    /**
     * Rewrite links if defined
     *
     * @param array $links List with links schema [url] and [desc] for description.
     *
     * @return array $links Return processed links.
     */
    protected function rewriteLinks($links = [])
    {
        // if configuration empty return unprocessed
        if (!isset($this->config->LinksRewrite)
            || count($this->config->LinksRewrite) == 0
        ) {
            return $links;
        }

        // if links list empty return unprocessed
        if (count($links) == 0) {
            return $links;
        }

        foreach ($links as &$link) {
            $link = $this->rewriteLink($link);
        }

        return array_filter($links);
    }

    /**
     * Rewrite link
     *
     * @param string $link Link to rewrite
     *
     * @return string $link Return processed link.
     *
     * @to-do resolve non consistently behaviour between method pattern and remove
     *        to handle slashes in urls
     */
    protected function rewriteLink(&$link)
    {
        $rewrite = $this->config->LinksRewrite->toArray();

        // helper function to replace the link description with a configured value
        $overwriteDescription = function (&$link, $currentRewrite) {
            // if another link description is configured use this instead
            if (isset($currentRewrite['description'])) {
                $link['desc'] = $currentRewrite['description'];
            }
        };

        foreach ($rewrite as $r) {
            // is remove pattern than suppress link refs #10834
            if (isset($r['remove'])) {
                if (0 != preg_match(
                    '/' . addcslashes($r['remove'], '/') . '/i',
                    trim($link['url'])
                )
                ) {
                    unset($link);
                    return;
                }
            }
            // is pattern set and matches so try rewrite url
            if (isset($r['pattern'])
                && 0 != preg_match('/' . $r['pattern'] . '/i', trim($link['url']))
            ) {
                // if function is set and available then perform on url
                if (isset($r['function']) && is_callable($r['function'])) {
                    $link['url'] = $r['function']($link['url']);
                }

                // is search and replace set so try to rewrite url
                if (isset($r['search']) && isset($r['replace'])) {
                    // prepare search pattern
                    // should be free of conflicting meta characters
                    $pattern = str_replace(['.'], ['\.'], $r['search']);
                    $pattern = '/(' . $pattern . ')/i';
                    // replace it only one time
                    $link['url'] = preg_replace(
                        $pattern,
                        trim($r['replace']),
                        trim($link['url']),
                        1,
                        $count
                    );
                    if ($count) {
                        // overwrite description as we have a successful replace
                        $overwriteDescription($link, $r);
                    }
                    // add http if needed
                    // @to-do make it https compatible
                    if (!preg_match('/^(https?:\/\/)/', $link['url'])) {
                        $link['url'] = 'https://' . $link['url'];
                    }
                }
                // is method set so call alternatively method proceed link
                if (isset($r['method']) && method_exists($this, $r['method'])) {
                    $link['url'] = $this->$r['method']($link['url']);
                    // overwrite description as we called the defined method
                    $overwriteDescription($link, $r);
                } // end if isset method
            } // end if isset pattern
        } // end foreach
        return $link;
    }

    /**
     * Resolve Rewrite/Schweitzer Url
     *
     * @param string $link Link to rewrite
     *
     * @return string $link Return processed link.
     */
    protected function resolveEblLink($link)
    {
        if (false === ($user = $this->manager->isLoggedIn())) {
            $id = $this->driver->getUniqueId();
            return $this->url->__invoke(
                'record-ebllink',
                [],
                ['query' => ['link' => $link, 'id' => $id]]
            );
        }
        $url = $this->rewrite->resolveLink($link, $user);
        return $url;
    }

    /**
     * Customized method for multi resolver support
     *
     * Get all the links associated with this record depending on the OpenURL setting
     * replace_other_urls.  Returns an array of associative arrays each containing
     * 'desc' and 'url' keys.
     *
     * @return bool
     */
    protected function hasOpenUrlReplaceSetting()
    {
        return isset($this->resolverConfig->General->replace_other_urls)
        && $this->resolverConfig->General->replace_other_urls;
    }

    /**
     * Get HTML to render a title.
     *
     * @param int $maxLength Maximum length of non-highlighted title.
     *
     * @return string
     */
    public function getTitleHtml($maxLength = 180)
    {
        $highlightedTitle = $this->driver->tryMethod('getHighlightedTitle');

        $titleRegexPattern = '/(\s[\/\.:]\s*)*$/';
        $shortTitle = preg_replace(
            $titleRegexPattern,
            '',
            trim($this->driver->tryMethod('getShortTitle'))
        );
        $subTitle = preg_replace(
            $titleRegexPattern,
            '',
            trim($this->driver->tryMethod('getSubTitle'))
        );
        $sectionTitle = preg_replace(
            $titleRegexPattern,
            '',
            trim($this->driver->tryMethod('getTitleSection'))
        );
        $title = (!empty($shortTitle) ? $shortTitle : preg_replace(
            $titleRegexPattern,
            '',
            trim($this->driver->tryMethod('getTitle'))
        )) .
        (!empty($subTitle) ? ': ' . $subTitle : '') .
        (!empty($sectionTitle) ? ', ' . $sectionTitle : '');

        if (!empty($highlightedTitle)) {
            $highlight = $this->getView()->plugin('highlight');
            $addEllipsis = $this->getView()->plugin('addEllipsis');
            return $highlight($addEllipsis($highlightedTitle, $title));
        }
        if (!empty($title)) {
            $escapeHtml = $this->getView()->plugin('escapeHtml');
            $truncate = $this->getView()->plugin('truncate');
            return $escapeHtml($truncate($title, $maxLength));
        }
        $transEsc = $this->getView()->plugin('transEsc');
        return $transEsc('Title not available');
    }

    /**
     * Get online editions
     *
     * @return array
     */
    public function getOnlineEditions()
    {
        $online_keys = [
            "Online-Ausg.",
            "Online-Ausg.:",
            "Online-Ausgabe",
            "Digital. Ausg.",
            "Online-Ausg. u.d.T.",
            "Elektronische Reproduktion"
        ];
        $onlineEditions = [];
        $recordLinkHelper = $this->getView()->plugin('recordLink');
        foreach ($this->driver->tryMethod('getAdditionals') as $add) {
            if (isset($add['identifier'])
                && (in_array($add['identifier'], $online_keys)
                || (isset($add['note'])
                && in_array($add['note'], $online_keys)))
            ) {
                if (isset($add['id'])) {
                    $link = $recordLinkHelper->getTabUrl($add['id'], 'Holdings');
                    if (!empty($link)) {
                        $add['link'] = $link;
                        $onlineEditions[] = $add;
                    }
                } elseif (isset($add['related_records'])) {
                    foreach (
                        $add['related_records']['first_results'] as $linkedRecord
                    ) {
                        $link = $recordLinkHelper->getTabUrl($linkedRecord);
                        if (!empty($link)) {
                            //overwrite link in entry and add to result array
                            $add['link'] = $link;
                            $onlineEditions[] = $add;
                        }
                    }
                }
            }
        }
        return $onlineEditions;
    }

    /**
     * Get advanced query params
     *
     * @return array
     */
    public function getAdvancedSearchQueryParams()
    {
        $queryParams = [];
        if ($title = $this->driver->getShortTitle()) {
            $queryParams['lookfor0'][] = $title;
            $queryParams['type0'][] = 'Title';
        }
        $isn = $this->driver->tryMethod('getCleanISSN');
        if (!empty($isn)) {
            $queryParams['lookfor0'][] = $isn;
            $queryParams['type0'][] = 'ISN';
        }
        if ($author = $this->driver->tryMethod('getPrimaryAuthors')) {
            $queryParams['lookfor0'][] = is_array($author) ?
              current($author) : $author;
            $queryParams['type0'][] = 'Author';
        }
        if (!empty($queryParams)) {
            // although this is a standard parameter
            // we should only pass it if there is an actual query to perform
            $queryParams['join'] = 'AND';
        }
        return $queryParams;
    }

    /**
     * Get the unique title id of the record
     *
     * @return string
     */
    public function getTitleDescribedById()
    {
        return "{$this->driver->getSourceIdentifier()}|{$this->driver->getUniqueId()}";
    }
}
