<?php
/**
 * Factory for record driver data formatting view helper
 *
 * PHP version 7
 *
 * Copyright (C) Villanova University 2016.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  View_Helpers
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:architecture:record_data_formatter
 * Wiki
 */
namespace finc\View\Helper\Root;

use VuFind\View\Helper\Root\RecordDataFormatter;

/**
 * Factory for record driver data formatting view helper
 *
 * @category VuFind
 * @package  View_Helpers
 * @author   Demian Katz <demian.katz@villanova.edu>
 * @author   Konsta Raunio <konsta.raunio@helsinki.fi>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:architecture:record_data_formatter
 * Wiki
 */
class RecordDataFormatterFactory
{

    use MultiDataFieldsTrait;

    /**
     * Create the helper.
     *
     * @return RecordDataFormatter
     */
    public function __invoke()
    {
        $helper = new RecordDataFormatter();

        $helper->setDefaults(
            'collection-info',
            [$this, 'getDefaultCollectionInfoSpecs']
        );
        $helper->setDefaults(
            'collection-record',
            [$this, 'getDefaultCollectionRecordSpecs']
        );
        $helper->setDefaults(
            'core',
            [$this, 'getDefaultCoreSpecs']
        );
        $helper->setDefaults(
            'description',
            [$this, 'getDefaultDescriptionSpecs']
        );
        $helper->setDefaults(
            'toc',
            [$this, 'getTOCSpecs']
        );
        $helper->setDefaults(
            'core-ai',
            [$this, 'getAiCoreSpecs']
        );
        $helper->setDefaults(
            'core-lido',
            [$this, 'getLidoCoreSpecs']
        );
        $helper->setDefaults(
            'description-lido',
            [$this, 'getLidoDescriptionSpecs']
        );
        $helper->setDefaults(
            'core-marc',
            [$this, 'getMarcCoreSpecs']
        );
        $helper->setDefaults(
            'topics-tab',
            [$this, 'getTopicsTabSpecs']
        );

        return $helper;
    }

    /**
     * Get default specifications for displaying data in core metadata.
     *
     * @return array
     */
    public function getAiCoreSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setTemplateLine(
            'Journal Title',
            'getContainerTitle',
            'data-containerTitle.phtml'
        );
        $spec->setTemplateLine(
            'Authors/Corporations',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'context' => [
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ]
            ]
        );
        $spec->setLine(
            'Edition',
            'getEdition',
            null,
            ['prefix' => '<span property="bookEdition">', 'suffix' => '</span>']
        );
        $spec->setTemplateLine(
            'In',
            'getJTitle',
            'data-jTitle.phtml'
        );
        $spec->setLine(
            'Format',
            'getFormats',
            'RecordHelper',
            ['helperMethod' => 'getFormatList']
        );
        $spec->setTemplateLine(
            'Language',
            'getLanguages',
            'data-transEscCommaSepLang.phtml'
        );
        $spec->setTemplateLine(
            'Published',
            'getPublicationDetails',
            'data-publicationDetails.phtml'
        );
        $spec->setLine(
            'Series',
            'getSeries',
            null,
            ['recordLink' => 'series']
        );
        $spec->setTemplateLine(
            'Subjects',
            'getAllSubjectHeadings',
            'data-allSubjectHeadings.phtml'
        );
        $spec->setTemplateLine(
            'child_records',
            'getChildRecordCount',
            'data-childRecords.phtml',
            ['allowZero' => false]
        );
        /* finc: remove 'Online Access' in record-details #13770 - VE */
        /* $spec->setTemplateLine('Online Access', true, 'data-onlineAccess.phtml'); */
        $spec->setTemplateLine(
            'Related Items',
            'getAllRecordLinks',
            'data-allRecordLinks.phtml'
        );
        $spec->setTemplateLine('Tags', true, 'data-tags.phtml');
        return $spec->getArray();
    }

    /**
     * Get default specifications for displaying data in collection-info metadata.
     *
     * @return array
     */
    public function getDefaultCollectionInfoSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setTemplateLine(
            'Authors/Corporations',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'context' => [
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ]
            ]
        );
        $spec->setLine(
            'Title',
            'getTitleDetails'
        );
        $spec->setLine(
            'Dates of publication',
            'getDateSpan'
        );
        $spec->setLine(
            'Abstract',
            'getSummary'
        );
        $spec->setTemplateLine(
            'Published',
            'getPublicationDetails',
            'data-publicationDetails.phtml'
        );
        $spec->setLine(
            'Edition',
            'getEdition',
            null,
            ['prefix' => '<span property="bookEdition">', 'suffix' => '</span>']
        );
        $spec->setTemplateLine(
            'Subjects',
            'getAllSubjectHeadings',
            'data-allSubjectHeadings.phtml'
        );
        /* finc: remove 'Online Access' in record-details #13770 - VE */
        /* $spec->setTemplateLine('Online Access', true, 'data-onlineAccess.phtml'); */
        $spec->setLine(
            'Item Description',
            'getGeneralNotes'
        );
        $spec->setLine(
            'Production Credits',
            'getProductionCredits'
        );
        $spec->setTemplateLine(
            'Set Multipart',
            'getHierarchyParentTitle',
            'data-hierarchyParentTitle.phtml'
        );
        $spec->setTemplateLine(
            'ISBN',
            'getISBNs',
            'data-isbn.phtml'
        );
        $spec->setTemplateLine(
            'ISSN',
            'getISSNs',
            'data-issn.phtml'
        );
        $spec->setLine(
            'Notes',
            'getAdditionalNotes',
            null,
            [
                'useCache' => true
            ]
        );
        $spec->setLine(
            'Format',
            'getFormats',
            'RecordHelper',
            ['helperMethod' => 'getFormatList']
        );
        $spec->setTemplateLine(
            'Language',
            'getLanguages',
            'data-transEscCommaSepLang.phtml'
        );
        $spec->setMultiLine(
            'Additionals',
            'getAdditionals',
            [$this,'additionals']
        );
        return $spec->getArray();
    }

    /**
     * Get default specifications for displaying data in collection-record metadata.
     *
     * @return array
     */
    public function getDefaultCollectionRecordSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setLine('Abstract', 'getSummary');
        $spec->setTemplateLine(
            'Authors/Corporations',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'context' => [
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ]
            ]
        );
        $spec->setTemplateLine(
            'Language',
            'getLanguages',
            'data-transEscCommaSepLang.phtml'
        );
        $spec->setLine(
            'Format',
            'getFormats',
            'RecordHelper',
            ['helperMethod' => 'getFormatList']
        );
        $spec->setLine(
            'Access',
            'getAccessRestrictions'
        );
        $spec->setLine(
            'Related Items',
            'getRelationshipNotes'
        );
        return $spec->getArray();
    }

    /**
     * Get default specifications for displaying data in core metadata.
     *
     * @return array
     */
    public function getDefaultCoreSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setTemplateLine(
            'Published in',
            'getContainerTitle',
            'data-containerTitle.phtml'
        );
        $spec->setLine(
            'New Title',
            'getNewerTitles',
            null,
            ['recordLink' => 'title']
        );
        $spec->setLine(
            'Previous Title',
            'getPreviousTitles',
            null,
            ['recordLink' => 'title']
        );
        $spec->setTemplateLine(
            'Authors/Corporations',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'context' => [
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ]
            ]
        );
        $spec->setLine(
            'Title',
            'getTitleDetails'
        );
        $spec->setTemplateLine(
            'Title Uniform',
            'getTitleUniform',
            'data-titleUniform.phtml',
            [
                'labelFunction' =>
                    [$this,'titleUniformLabel']
            ]
        );
        $spec->setLine(
            'Edition',
            'getEdition',
            null,
            ['prefix' => '<span property="bookEdition">', 'suffix' => '</span>']
        );
        $spec->setLine(
            'Dissertation Note',
            'getDissertationNote',
            null,
            [
                'separator' => ', '
            ]
        );
        $spec->setLine(
            'Format',
            'getFormats',
            'RecordHelper',
            ['helperMethod' => 'getFormatList']
        );
        $spec->setTemplateLine(
            'Language',
            'getLanguages',
            'data-transEscCommaSepLang.phtml'
        );
        $spec->setMultiLine(
            'Published',
            true,
            [$this,'publicationDetailsImprint']
        );
        $spec->setTemplateLine(
            'Set Multipart',
            'getHierarchyParentTitle',
            'data-hierarchyParentTitle.phtml'
        );
        $spec->setTemplateLine(
            'Subjects',
            'getAllSubjectHeadings',
            'data-allSubjectHeadings.phtml'
        );
        $spec->setTemplateLine(
            'child_records',
            'getChildRecordCount',
            'data-childRecords.phtml',
            ['allowZero' => false]
        );
        /* finc: remove 'Online Access' in record-details #13770 - VE */
        /* $spec->setTemplateLine('Online Access', true, 'data-onlineAccess.phtml'); */
        $spec->setTemplateLine(
            'Related Items',
            'getAllRecordLinks',
            'data-allRecordLinks.phtml'
        );
        $spec->setMultiLine(
            'Additionals',
            'getAdditionals',
            [$this,'additionals']
        );
        $spec->setLine(
            'Source',
            'getMegaCollection'
        );
        $spec->setMultiLine(
            "Other Relationship Entry",
            'getOtherRelationshipEntry',
            [$this,'otherRelationshipEntry']
        );
        $spec->setLine(
            'Notes',
            'getAdditionalNotes',
            null,
            [
                'useCache' => true
            ]
        );
        $spec->setTemplateLine('Tags', true, 'data-tags.phtml');
        return $spec->getArray();
    }

    /**
     * Get default specifications for displaying data in the description tab.
     *
     * @return array
     */
    public function getDefaultDescriptionSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setLine(
            'Abstract',
            'getSummary'
        );
        $spec->setLine(
            'Dates of publication',
            'getDateSpan'
        );
        $spec->setLine(
            'Item Description',
            'getGeneralNotes'
        );
        $spec->setLine(
            'Physical Description',
            'getPhysicalDescriptions'
        );
        $spec->setLine(
            'Publication Frequency',
            'getPublicationFrequency'
        );
        $spec->setLine(
            'Playing Time',
            'getPlayingTimes'
        );
        $spec->setLine(
            'Format',
            'getSystemDetails'
        );
        $spec->setLine(
            'Audience',
            'getTargetAudienceNotes'
        );
        $spec->setLine(
            'Awards',
            'getAwards'
        );
        $spec->setLine(
            'Production Credits',
            'getProductionCredits'
        );
        $spec->setLine(
            'Bibliography',
            'getBibliographyNotes'
        );
        $spec->setTemplateLine(
            'ISBN',
            'getISBNs',
            'data-isbn.phtml'
        );
        $spec->setTemplateLine(
            'ISSN',
            'getISSNs',
            'data-issn.phtml'
        );
        $spec->setLine(
            'DOI',
            'getCleanDOI'
        );
        $spec->setLine(
            'EISSN',
            'getEISSNs'
        );
        $spec->setLine(
            'Related Items',
            'getRelationshipNotes'
        );
        $spec->setLine(
            'Access',
            'getAccessRestrictions'
        );
        $spec->setLine(
            'Finding Aid',
            'getFindingAids'
        );
        $spec->setLine(
            'Publication_Place',
            'getHierarchicalPlaceNames'
        );
        $spec->setTemplateLine(
            'Author Notes',
            true,
            'data-authorNotes.phtml'
        );
        $spec->setTemplateLine(
            'Call Number',
            'getLocalSignature',
            'data-localSignature.phtml'
        );
        $spec->setLine(
            'Notes',
            'getAdditionalNotes',
            null,
            [
                'useCache' => true
            ]
        );
        return $spec->getArray();
    }

    /**
     * Get default specifications for displaying data in collection-info metadata.
     *
     * @return array
     */
    public function getLidoCoreSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setTemplateLine(
            'Originators',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'labelFunction' => function ($data) {
                    return count($data['primary']) > 1
                        ? 'Originators' : 'Originator';
                },
                'context' => [
                    'type' => 'primary',
                    'schemaLabel' => 'author',
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ],
            ]
        );
        $spec->setTemplateLine(
            'Corporate Originator',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'labelFunction' => function ($data) {
                    return count($data['corporate']) > 1
                        ? 'Corporate Originators' : 'Corporate Originator';
                },
                'context' => [
                    'type' => 'corporate',
                    'schemaLabel' => 'creator',
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ],
            ]
        );
        $spec->setTemplateLine(
            'Other Originators',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'context' => [
                    'type' => 'secondary',
                    'schemaLabel' => 'contributor',
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ],
            ]
        );
        $spec->setLine(
            'Subject Detail',
            'getSubjectDetails'
        );
        $spec->setLine(
            'Subject Place',
            'getSubjectPlaces'
        );
        $spec->setLine(
            'Subject Date',
            'getSubjectDates'
        );

        $spec->setLine(
            'Subject Actor',
            'getSubjectActors'
        );
        $spec->setLine(
            'Institution',
            'getInstitutions'
        );
        $spec->setLine(
            'Inventory ID',
            'getIdentifier'
        );
        $spec->setLine(
            'Inventory ID',
            'getIdentifier'
        );
        $spec->setLine(
            'Measurements',
            'getMeasurements'
        );
        $spec->setLine(
            'Measurements',
            'getMeasurementsDescription'
        );
        $spec->setLine(
            'Collection',
            'getCollections'
        );
        $spec->setLine(
            'Object type',
            'getFormats',
            'RecordHelper',
            ['helperMethod' => 'getFormatList']
        );
        $spec->setLine(
            'Other Classification',
            'getFormatClassifications'
        );
        $spec->setLine(
            'Other ID',
            'getLocalIdentifiers'
        );
        $spec->setMultiLine(
            'Events',
            'getEvents',
            [$this,'events']
        );
        // , context: "recordEvents"
        $spec->setTemplateLine(
            'Language',
            'getLanguages',
            'data-transEscCommaSepLang.phtml'
        );
        $spec->setLine(
            'Time of origin',
            'getDateSpan',
            'recordHelper',
            [
                'helperMethod' => 'mapDateListToRangeView'
            ]
        );
        $spec->setLine(
            'Edition',
            'getEdition',
            null,
            [
                'prefix' => '<span property="bookEdition">',
                'suffix' => '</span>'
            ]
        );
        $spec->setTemplateLine(
            'Subjects',
            'getAllSubjectHeadings',
            'data-allSubjectHeadings.phtml'
        );
        /* finc: remove 'Online Access' in record-details #13770 -VE */
        /* $spec->setTemplateLine('Online Access', true, 'data-onlineAccess.phtml'); */
        $spec->setTemplateLine(
            'Related Items',
            'getAllRecordLinks',
            'data-allRecordLinks.phtml'
        );
        return $spec->getArray();
    }

    /**
     * Get default specifications for displaying data in collection-record metadata.
     *
     * @return array
     */
    public function getLidoDescriptionSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setLine(
            'Time of origin',
            'getDateSpan',
            'recordHelper',
            [
                'helperMethod' => 'mapDateListToRangeView'
            ]
        );
        $spec->setTemplateLine(
            'Access',
            'getAccessNote',
            'data-accessNote.phtml'
        );
        return $spec->getArray();
    }

    /**
     * Get marc specifications for displaying data in core metadata.
     *
     * @return array
     */
    public function getMarcCoreSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setTemplateLine(
            'Published in',
            'getContainerTitle',
            'data-containerTitle.phtml'
        );
        $spec->setLine(
            'New Title',
            'getNewerTitles',
            'data-linkViaFincId.phtml'
        );
        $spec->setLine(
            'Previous Title',
            'getPreviousTitles',
            'data-linkViaFincId.phtml'
        );
        $spec->setTemplateLine(
            'Authors/Corporations',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'context' => [
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ]
            ]
        );
        $spec->setLine(
            'Title',
            'getTitleDetails',
            null,
            ['separator' => '<br />']
        );
        $spec->setTemplateLine(
            'Title Uniform',
            'getTitleUniform',
            'data-titleUniform.phtml',
            [
                'labelFunction' =>
                    [$this,'titleUniformLabel']
            ]
        );
        $spec->setLine(
            'Edition',
            'getEdition',
            null,
            ['prefix' => '<span property="bookEdition">', 'suffix' => '</span>']
        );
        $spec->setLine(
            'Dissertation Note',
            'getDissertationNote',
            null,
            [
                'separator' => ', '
            ]
        );
        $spec->setLine(
            'Format',
            'getFormats',
            'RecordHelper',
            ['helperMethod' => 'getFormatList']
        );
        $spec->setTemplateLine(
            'Language',
            'getLanguages',
            'data-transEscCommaSepLang.phtml'
        );
        $spec->setTemplateLine(
            'Published',
            'getPublicationDetails',
            'data-publicationDetails.phtml'
        );
        $spec->setTemplateLine(
            'German Prints Index Number',
            'getIndexOfGermanPrints',
            'data-indexOfGermanPrints.phtml'
        );
        $spec->setTemplateLine(
            'Set Multipart',
            'getHierarchyParentTitle',
            'data-hierarchyParentTitle.phtml'
        );
        $spec->setTemplateLine(
            'Subjects',
            'getAllSubjectHeadings',
            'data-allSubjectHeadings.phtml'
        );
        $spec->setTemplateLine(
            'child_records',
            'getChildRecordCount',
            'data-childRecords.phtml',
            ['allowZero' => false]
        );
        $spec->setMultiLine(
            'Additionals',
            'getAdditionals',
            [$this,'additionals']
        );
        /* finc: remove 'Online Access' in record-details #13770 - VE */
        /* $spec->setTemplateLine('Online Access', true, 'data-onlineAccess.phtml'); */
        $spec->setTemplateLine(
            'Related Items',
            'getAllRecordLinks',
            'data-allRecordLinks.phtml'
        );
        $spec->setTemplateLine(
            'Local Subjects',
            'getLocalSubjects',
            'data-localSubjects.phtml',
            [
                'labelFunction' => function () {
                    return 'Subject';
                }
            ]
        );
        $spec->setLine(
            'Source',
            'getMegaCollection'
        );
        $spec->setMultiLine(
            "Other Relationship Entry",
            'getOtherRelationshipEntry',
            [$this,'otherRelationshipEntry']
        );
        $spec->setLine(
            'Notes',
            'getAdditionalNotes',
            null,
            [
                'useCache' => true
            ]
        );
        $spec->setTemplateLine('Tags', true, 'data-tags.phtml');
        return $spec->getArray();
    }

    /**
     * Labels titleUniform Data Line based on the record
     * being catalogued following RDA, or not. cf. https://projekte.ub.uni-leipzig.de/issues/13830
     * @param $data
     * @param $driver
     * @return string
     */
    public function titleUniformLabel($data, $driver)
    {
        return $driver->tryMethod('isRDA')
            ? 'rda_original_title'
            : 'non_rda_original_title';
    }

    /**
     * Get marc specifications for displaying table of contents.
     *
     * @return array
     */
    public function getTOCSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setTemplateLine(
            'Table of Contents',
            'getTOC',
            'data-toc.phtml'
        );
        return $spec->getArray();
    }
    
    /**
     * Get default specifications for displaying data in the topics tab.
     *
     * @return array
     */
    public function getTopicsTabSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setTemplateLine(
            'RVK',
            'getRvkWithMetadata',
            'data-rvkNotation.phtml'
        );
        $spec->setTemplateLine(
            'Subjects',
            'getTopics',
            'data-allSubjectHeadingsWithExtended.phtml'
        );
        return $spec->getArray();
    }

}
