<?php
/**
 * PHP version 7
 *
 * Copyright (C) Villanova University 2022.
 * Copyright (C) Leipzig University Library 2022.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  finc
 * @author   Robert Lange  <lange@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 */
namespace fincTest\RecordDriver;

use finc\RecordDriver\SolrAI as SolrAIDriver;
use finc\RecordDriver\SolrDefault;
use finc\RecordDriver\SolrMarcFinc as SolrMarcFincDriver;
use VuFindTest\Unit\TestCase as VuFindTestCase;

/**
 * SolrBaseTestCase - base helper methods and classes
 *
 * @category VuFind
 * @package  finc
 * @author   Robert Lange  <lange@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     http://vufind.org
 */
class SolrBaseTestCase extends VuFindTestCase
{
    /**
     * SolrDefault
     *
     * @var SolrDefault
     */
    protected $driver;

    /**
     * Initialize driver with fixture
     *
     * @param String $file File
     *
     * @return void
     */
    public function initialize($file)
    {
        if (!$this->driver) {
            $this->driver = new SolrMarcFincDriver();
            $fixture = $this->getFixtureData($file);
            $this->driver->setRawData($fixture);
        }
    }

    /**
     * Get record fixture
     *
     * @param String $file File
     *
     * @return array
     */
    protected function getFixtureData($file, $decode = true)
    {
        $content = file_get_contents(realpath(FINC_TEST_FIXTURES . '/' . $file));
        return $decode ? json_decode($content, true) : $content;
    }

    protected function setUp(): void
    {
        parent::setUp();
        $this->driver = $this->getDriverMock();
    }

    /**
     * Get a working renderer.
     *
     * @param array  $plugins Custom VuFind plug-ins to register
     * @param string $theme   Theme directory to load from
     *
     * @return \Laminas\View\Renderer\PhpRenderer
     */
    protected function getPhpRenderer($plugins = [], $theme = 'finc')
    {
        $resolver = new \Laminas\View\Resolver\TemplatePathStack();

        // This assumes that all themes will be testing inherit directly
        // from root with no intermediate themes.  Probably safe for most
        // test situations, though other scenarios are possible.
        $resolver->setPaths(
            [
                APPLICATION_PATH . '/themes/' . $theme . '/templates'
            ]
        );
        $renderer = new \Laminas\View\Renderer\PhpRenderer();
        $renderer->setResolver($resolver);
        if (!empty($plugins)) {
            $pluginManager = $renderer->getHelperPluginManager();
            foreach ($plugins as $key => $value) {
                if ($key == 'record') {
                    $value->setView($renderer);
                }
                $pluginManager->setService($key, $value);
            }
        }
        return $renderer;
    }

    /**
     * Get plugins to register to support view helper being tested
     *
     * @return array
     */
    protected function getPlugins($id)
    {
        $record = new RecordViewHelperMock($this->driver);
        $serverUrl = $this->createMock(\Laminas\View\Helper\ServerUrl::class);
        /*TODO: should be called only once*/
        $serverUrl->expects(/*$this->once()*/$this->any())->method('__invoke')
            ->will($this->returnValue('http://localhost:11100/'));

        $config = $this->createMock(\VuFind\View\Helper\Root\Config::class);
        $config->expects($this->any())->method('get')
            ->will($this->returnValue((object)['Site' => (object)['title' => 'Generic Katalog (devel)']]));

        $url = $this->createMock(\VuFind\View\Helper\Root\Url::class);
        /*TODO: should be called only once*/
        $url->expects(/*$this->once()*/$this->any())->method('__invoke')
            ->will($this->returnValue("Record/$id"));

        return [
            'record' => $record,
            'serverUrl' => $serverUrl,
            'config' => $config,
            'url'   => $url,
            'context' => new \VuFind\View\Helper\Root\Context(),
            'slot' => new \VuFindTheme\View\Helper\Slot()
        ];
    }
}

class RecordViewHelperMock extends \VuFind\View\Helper\Root\Record
{
    public function __construct($driver)
    {
        $this->driver = $driver;
    }
}

